// THIS IS A GENERATED FILE. DO NOT EDIT. SOURCE metrics.py AND metrics_table.py
{"questions", MetricDefinition(PrometheusMetricType::counter, "Number of questions")},
{"ipv6-questions", MetricDefinition(PrometheusMetricType::counter, "Number of IPv6 questions")},
{"tcp-questions", MetricDefinition(PrometheusMetricType::counter, "Number of TCP questions")},
{"cache-hits", MetricDefinition(PrometheusMetricType::counter, "Number of cache hits")},
{"cache-misses", MetricDefinition(PrometheusMetricType::counter, "Number of cache misses")},
{"cache-entries", MetricDefinition(PrometheusMetricType::gauge, "Number of record cache entries")},
{"max-cache-entries", MetricDefinition(PrometheusMetricType::gauge, "Currently configured maximum number of cache entries")},
{"max-packetcache-entries", MetricDefinition(PrometheusMetricType::gauge, "Currently configured maximum number of packet cache entries")},
{"cache-bytes", MetricDefinition(PrometheusMetricType::gauge, "Size of the cache in bytes")},
{"packetcache-hits", MetricDefinition(PrometheusMetricType::counter, "Number of packetcache hits")},
{"packetcache-misses", MetricDefinition(PrometheusMetricType::counter, "Number of packetcache misses")},
{"packetcache-entries", MetricDefinition(PrometheusMetricType::gauge, "Number of packetcache entries")},
{"packetcache-bytes", MetricDefinition(PrometheusMetricType::gauge, "Size of the packetcache in bytes")},
{"malloc-bytes", MetricDefinition(PrometheusMetricType::gauge, "Number of bytes allocated by malloc")},
{"servfail-answers", MetricDefinition(PrometheusMetricType::counter, "Number of servfail answers")},
{"nxdomain-answers", MetricDefinition(PrometheusMetricType::counter, "Number of nxdomain answers")},
{"noerror-answers", MetricDefinition(PrometheusMetricType::counter, "Number of noerror answers")},
{"unauthorized-udp", MetricDefinition(PrometheusMetricType::counter, "Number of unauthorized UDP queries")},
{"unauthorized-tcp", MetricDefinition(PrometheusMetricType::counter, "Number of unauthorized TCP queries")},
{"tcp-client-overflow", MetricDefinition(PrometheusMetricType::counter, "Number of TCP client connections refused because of too many connections")},
{"client-parse-errors", MetricDefinition(PrometheusMetricType::counter, "Number of client parse errors")},
{"server-parse-errors", MetricDefinition(PrometheusMetricType::counter, "Number of server parse errors")},
{"too-old-drops", MetricDefinition(PrometheusMetricType::counter, "Number of queries dropped because of a timeout")},
{"answers0-1", MetricDefinition(PrometheusMetricType::counter, "Number of queries answered in less than 1 ms")},
{"answers1-10", MetricDefinition(PrometheusMetricType::counter, "Number of queries answered in 1-10 ms")},
{"answers10-100", MetricDefinition(PrometheusMetricType::counter, "Number of queries answered in 10-100 ms")},
{"answers100-1000", MetricDefinition(PrometheusMetricType::counter, "Number of queries answered in 100-1000 ms")},
{"answers-slow", MetricDefinition(PrometheusMetricType::counter, "Number of queries answered in more than 1000 ms")},
{"x-ourtime0-1", MetricDefinition(PrometheusMetricType::counter, "Counts responses where between 0 and 1 milliseconds was spent within the Recursor")},
{"x-ourtime1-2", MetricDefinition(PrometheusMetricType::counter, "Counts responses where between 1 and 2 milliseconds was spent within the Recursor")},
{"x-ourtime2-4", MetricDefinition(PrometheusMetricType::counter, "Counts responses where between 16 and 32 milliseconds was spent within the Recursor")},
{"x-ourtime4-8", MetricDefinition(PrometheusMetricType::counter, "Counts responses where between 4 and 8 milliseconds was spent within the Recursor")},
{"x-ourtime8-16", MetricDefinition(PrometheusMetricType::counter, "Counts responses where between 8 and 16 milliseconds was spent within the Recursor")},
{"x-ourtime16-32", MetricDefinition(PrometheusMetricType::counter, "Counts responses where between 16 and 32 milliseconds was spent within the Recursor")},
{"x-ourtime-slow", MetricDefinition(PrometheusMetricType::counter, "Counts responses where more than 32 milliseconds was spent within the Recursor")},
{"auth4-answers0-1", MetricDefinition(PrometheusMetricType::counter, "Number of IPv4 queries answered in less than 1 ms")},
{"auth4-answers1-10", MetricDefinition(PrometheusMetricType::counter, "Number of IPv4 queries answered in 1-10 ms")},
{"auth4-answers10-100", MetricDefinition(PrometheusMetricType::counter, "Number of IPv4 queries answered in 10-100 ms")},
{"auth4-answers100-1000", MetricDefinition(PrometheusMetricType::counter, "Number of IPv4 queries answered in 100-1000 ms")},
{"auth4-answers-slow", MetricDefinition(PrometheusMetricType::counter, "Number of IPv4 queries answered in more than 1000 ms")},
{"auth6-answers0-1", MetricDefinition(PrometheusMetricType::counter, "Number of IPv6 queries answered in less than 1 ms")},
{"auth6-answers1-10", MetricDefinition(PrometheusMetricType::counter, "Number of IPv6 queries answered in 1-10 ms")},
{"auth6-answers10-100", MetricDefinition(PrometheusMetricType::counter, "Number of IPv6 queries answered in 10-100 ms")},
{"auth6-answers100-1000", MetricDefinition(PrometheusMetricType::counter, "Number of IPv6 queries answered in 100-1000 ms")},
{"auth6-answers-slow", MetricDefinition(PrometheusMetricType::counter, "Number of IPv6 queries answered in more than 1000 ms")},
{"qa-latency", MetricDefinition(PrometheusMetricType::gauge, "Shows the current latency average, in microseconds, exponentially weighted over past 'latency-statistic-size' packets")},
{"x-our-latency", MetricDefinition(PrometheusMetricType::gauge, "Shows the averaged time spent within PowerDNS, in microseconds, exponentially weighted over past 'latency-statistic-size' packets")},
{"unexpected-packets", MetricDefinition(PrometheusMetricType::counter, "Number of unexpected packets")},
{"case-mismatches", MetricDefinition(PrometheusMetricType::counter, "Number of case mismatches")},
{"spoof-prevents", MetricDefinition(PrometheusMetricType::counter, "Number of spoof prevents")},
{"nsset-invalidations", MetricDefinition(PrometheusMetricType::counter, "Number of nsset invalidations")},
{"resource-limits", MetricDefinition(PrometheusMetricType::counter, "Number of resolution aborted because of a local resource limit")},
{"over-capacity-drops", MetricDefinition(PrometheusMetricType::counter, "Number of queries dropped because the threads limit was reached")},
{"policy-drops", MetricDefinition(PrometheusMetricType::counter, "Number of queries dropped because of a policy")},
{"no-packet-error", MetricDefinition(PrometheusMetricType::counter, "Number of calls to recvmsg() that returned no packet even though the socket was ready")},
{"dlg-only-drops", MetricDefinition(PrometheusMetricType::counter, "Obsolete")},
{"ignored-packets", MetricDefinition(PrometheusMetricType::counter, "Number of ignored packets")},
{"max-mthread-stack", MetricDefinition(PrometheusMetricType::counter, "Maximum amount of the mthread stack ever used")},
{"negcache-entries", MetricDefinition(PrometheusMetricType::gauge, "Number of negcache entries")},
{"throttle-entries", MetricDefinition(PrometheusMetricType::gauge, "Number of throttle entries")},
{"nsspeeds-entries", MetricDefinition(PrometheusMetricType::gauge, "Number of nsspeeds entries")},
{"failed-host-entries", MetricDefinition(PrometheusMetricType::gauge, "Number of entries in the failed NS cache")},
{"concurrent-queries", MetricDefinition(PrometheusMetricType::gauge, "Number of concurrent queries")},
{"security-status", MetricDefinition(PrometheusMetricType::gauge, "Current security status")},
{"outgoing-timeouts", MetricDefinition(PrometheusMetricType::counter, "Number of outgoing timeouts")},
{"outgoing4-timeouts", MetricDefinition(PrometheusMetricType::counter, "Number of IPv4 outgoing timeouts")},
{"outgoing6-timeouts", MetricDefinition(PrometheusMetricType::counter, "Number of IPv6 outgoing timeouts")},
{"auth-zone-queries", MetricDefinition(PrometheusMetricType::counter, "Number of queries to locally hosted authoritative zones ('setting-auth-zones')")},
{"tcp-outqueries", MetricDefinition(PrometheusMetricType::counter, "Number of outgoing TCP queries sent")},
{"all-outqueries", MetricDefinition(PrometheusMetricType::counter, "Number of outgoing queries")},
{"ipv6-outqueries", MetricDefinition(PrometheusMetricType::counter, "Number of IPv6 outgoing queries sent")},
{"throttled-outqueries", MetricDefinition(PrometheusMetricType::counter, "Number of throttled outgoing queries")},
{"dont-outqueries", MetricDefinition(PrometheusMetricType::counter, "Number of outgoing queries not sent because of a 'dont-query' setting")},
{"throttled-out", MetricDefinition(PrometheusMetricType::counter, "Number of throttled outgoing queries")},
{"unreachables", MetricDefinition(PrometheusMetricType::counter, "Number of errors due to an unreachable server")},
{"ecs-queries", MetricDefinition(PrometheusMetricType::counter, "Number of outgoing queries adorned with an EDNS Client Subnet option")},
{"ecs-responses", MetricDefinition(PrometheusMetricType::counter, "Number of responses received from authoritative servers with an EDNS Client Subnet option we used")},
{"chain-resends", MetricDefinition(PrometheusMetricType::counter, "Number of chain resends")},
{"tcp-clients", MetricDefinition(PrometheusMetricType::gauge, "Number of TCP clients")},
{"udp-recvbuf-errors", MetricDefinition(PrometheusMetricType::counter, "Number of UDP recvbuf errors (Linux only)")},
{"udp-sndbuf-errors", MetricDefinition(PrometheusMetricType::counter, "Number of UDP sndbuf errors (Linux only)")},
{"udp-noport-errors", MetricDefinition(PrometheusMetricType::counter, "Number of UDP noport errors (Linux only)")},
{"udp-in-errors", MetricDefinition(PrometheusMetricType::counter, "Number of UDP in errors (Linux only)")},
{"edns-ping-matches", MetricDefinition(PrometheusMetricType::counter, "Number of EDNS Ping matches")},
{"edns-ping-mismatches", MetricDefinition(PrometheusMetricType::counter, "Number of EDNS Ping mismatches")},
{"dnssec-queries", MetricDefinition(PrometheusMetricType::counter, "Number of DNSSEC queries")},
{"noping-outqueries", MetricDefinition(PrometheusMetricType::counter, "Number of outgoing queries without ping")},
{"noedns-outqueries", MetricDefinition(PrometheusMetricType::counter, "Number of outgoing queries without EDNS")},
{"uptime", MetricDefinition(PrometheusMetricType::counter, "Process uptime in seconds")},
{"real-memory-usage", MetricDefinition(PrometheusMetricType::gauge, "Memory usage")},
{"fd-usage", MetricDefinition(PrometheusMetricType::gauge, "File descriptors usage")},
{"user-msec", MetricDefinition(PrometheusMetricType::counter, "CPU usage (user) in ms")},
{"sys-msec", MetricDefinition(PrometheusMetricType::counter, "CPU usage (system) in ms")},
{"dnssec-validations", MetricDefinition(PrometheusMetricType::counter, "Number of responses sent, packet-cache hits excluded, for which a DNSSEC validation was requested by either the client or the configuration")},
{"dnssec-result-insecure", MetricDefinition(PrometheusMetricType::counter, "Number of responses sent, excluding packet-cache hits, that were in the DNSSEC insecure state")},
{"dnssec-result-secure", MetricDefinition(PrometheusMetricType::counter, "Number of responses sent, excluding packet-cache hits, that were in the DNSSEC secure state")},
{"dnssec-result-bogus", MetricDefinition(PrometheusMetricType::counter, "Number of responses sent, excluding packet-cache hits, that were in the DNSSEC bogus state")},
{"dnssec-result-indeterminate", MetricDefinition(PrometheusMetricType::counter, "Number of responses sent, excluding packet-cache hits, that were in the DNSSEC indeterminate state")},
{"dnssec-result-nta", MetricDefinition(PrometheusMetricType::counter, "Number of responses sent, excluding packet-cache hits, that were in the DNSSEC NTA state")},
{"policy-result-noaction", MetricDefinition(PrometheusMetricType::counter, "Number of policy-mandated no-action results")},
{"policy-result-drop", MetricDefinition(PrometheusMetricType::counter, "Number of policy-mandated drops")},
{"policy-result-nxdomain", MetricDefinition(PrometheusMetricType::counter, "Number of policy-mandated NXdomain results")},
{"policy-result-nodata", MetricDefinition(PrometheusMetricType::counter, "Number of policy-mandated nodata results")},
{"policy-result-truncate", MetricDefinition(PrometheusMetricType::counter, "Number of policy-mandated truncate results")},
{"policy-result-custom", MetricDefinition(PrometheusMetricType::counter, "Number of policy-mandated custom results")},
{"query-pipe-full-drops", MetricDefinition(PrometheusMetricType::counter, "Number of queries dropped because the query distribution pipe was full")},
{"truncated-drops", MetricDefinition(PrometheusMetricType::counter, "Number of queries dropped because they were larger than 512 bytes")},
{"empty-queries", MetricDefinition(PrometheusMetricType::counter, "Number of queries dropped because they had a QD count of 0")},
{"dnssec-authentic-data-queries", MetricDefinition(PrometheusMetricType::counter, "Number of queries received with the AD bit set")},
{"dnssec-check-disabled-queries", MetricDefinition(PrometheusMetricType::counter, "Number of queries received with the CD bit set")},
{"variable-responses", MetricDefinition(PrometheusMetricType::counter, "Number of variable responses")},
{"special-memory-usage", MetricDefinition(PrometheusMetricType::gauge, "Memory usage (more precise but expensive to retrieve)")},
{"rebalanced-queries", MetricDefinition(PrometheusMetricType::counter, "Number of queries re-distributed because the first selected worker thread was above the target load")},
{"qname-min-fallback-success", MetricDefinition(PrometheusMetricType::counter, "Number of successful queries due to fallback mechanism within 'qname-minimization' setting")},
{"proxy-protocol-invalid", MetricDefinition(PrometheusMetricType::counter, "Number of invalid proxy protocol headers received")},
{"record-cache-contended", MetricDefinition(PrometheusMetricType::counter, "Number of contended record cache lock acquisitions")},
{"record-cache-acquired", MetricDefinition(PrometheusMetricType::counter, "Number of record cache lock acquisitions")},
{"nod-lookups-dropped-oversize", MetricDefinition(PrometheusMetricType::counter, "Number of NOD lookups dropped because they would exceed the maximum name length")},
{"taskqueue-pushed", MetricDefinition(PrometheusMetricType::counter, "Number of tasks pushed to the taskqueues")},
{"taskqueue-expired", MetricDefinition(PrometheusMetricType::counter, "Number of tasks expired before they could be run")},
{"taskqueue-size", MetricDefinition(PrometheusMetricType::gauge, "Number of tasks currently in the taskqueues")},
{"aggressive-nsec-cache-entries", MetricDefinition(PrometheusMetricType::counter, "Number of entries in the aggressive NSEC cache")},
{"aggressive-nsec-cache-nsec-hits", MetricDefinition(PrometheusMetricType::counter, "Number of NSEC-related hits from the aggressive NSEC cache")},
{"aggressive-nsec-cache-nsec3-hits", MetricDefinition(PrometheusMetricType::counter, "Number of NSEC3-related hits from the aggressive NSEC cache")},
{"aggressive-nsec-cache-nsec-wc-hits", MetricDefinition(PrometheusMetricType::counter, "Number of answers synthesized from the NSEC aggressive cache")},
{"aggressive-nsec-cache-nsec3-wc-hits", MetricDefinition(PrometheusMetricType::counter, "Number of answers synthesized from the NSEC3 aggressive cache")},
{"dot-outqueries", MetricDefinition(PrometheusMetricType::counter, "Number of outgoing DoT queries")},
{"dns64-prefix-answers", MetricDefinition(PrometheusMetricType::counter, "Number of answers generated by dns64-prefix matching")},
{"almost-expired-pushed", MetricDefinition(PrometheusMetricType::counter, "Number of almost-expired tasks pushed")},
{"almost-expired-run", MetricDefinition(PrometheusMetricType::counter, "Number of almost-expired tasks run to completion")},
{"almost-expired-exceptions", MetricDefinition(PrometheusMetricType::counter, "Number of almost-expired tasks that caused an exception")},
{"udp-in-csum-errors", MetricDefinition(PrometheusMetricType::counter, "Number of UDP in checksum errors (Linux only)")},
{"udp6-recvbuf-errors", MetricDefinition(PrometheusMetricType::counter, "Number of UDP6 recvbuf errors (Linux only)")},
{"udp6-sndbuf-errors", MetricDefinition(PrometheusMetricType::counter, "Number of UDP6 sndbuf errors (Linux only)")},
{"udp6-noport-errors", MetricDefinition(PrometheusMetricType::counter, "Number of UDP6 noport errors (Linux only)")},
{"udp6-in-errors", MetricDefinition(PrometheusMetricType::counter, "Number of UDP6 in errors (Linux only)")},
{"udp6-in-csum-errors", MetricDefinition(PrometheusMetricType::counter, "Number of UDP6 in checksum errors (Linux only)")},
{"cpu-iowait", MetricDefinition(PrometheusMetricType::counter, "Time spent waiting for I/O to complete by the whole system, in units of USER_HZ.")},
{"cpu-steal", MetricDefinition(PrometheusMetricType::counter, "Stolen time, which is the time spent by the whole system in other operating systems when running in a virtualized environment, in units of USER_HZ")},
{"cpu-msec-thread-0", MetricDefinition(PrometheusMetricType::multicounter, "Number of milliseconds spent in thread n")},
{"memory-allocs", MetricDefinition(PrometheusMetricType::counter, "Only relevant for development and if malloc tracing is enabled")},
{"memory-alloc-flux", MetricDefinition(PrometheusMetricType::counter, "Only relevant for development and if malloc tracing is enabled")},
{"memory-allocated", MetricDefinition(PrometheusMetricType::counter, "Only relevant for development and if malloc tracing is enabled")},
{"dnssec-result-bogus-no-valid-dnskey", MetricDefinition(PrometheusMetricType::counter, "Number of responses sent, packet-cache hits excluded, that were in the Bogus state because a valid DNSKEY could not be found")},
{"dnssec-result-bogus-invalid-denial", MetricDefinition(PrometheusMetricType::counter, "Number of responses sent, packet-cache hits excluded, that were in the Bogus state because a valid denial of existence proof could not be found")},
{"dnssec-result-bogus-unable-to-get-dss", MetricDefinition(PrometheusMetricType::counter, "Number of responses sent, packet-cache hits excluded, that were in the Bogus state because a valid DS could not be retrieved")},
{"dnssec-result-bogus-unable-to-get-dnskeys", MetricDefinition(PrometheusMetricType::counter, "Number of responses sent, packet-cache hits excluded, that were in the Bogus state because a valid DNSKEY could not be retrieved")},
{"dnssec-result-bogus-self-signed-ds", MetricDefinition(PrometheusMetricType::counter, "Number of responses sent, packet-cache hits excluded, that were in the Bogus state because a DS record was signed by itself")},
{"dnssec-result-bogus-no-rrsig", MetricDefinition(PrometheusMetricType::counter, "Number of responses sent, packet-cache hits excluded, that were in the Bogus state because required RRSIG records were not present in an answer")},
{"dnssec-result-bogus-no-valid-rrsig", MetricDefinition(PrometheusMetricType::counter, "Number of responses sent, packet-cache hits excluded, that were in the Bogus state because only invalid RRSIG records were present in an answer")},
{"dnssec-result-bogus-missing-negative-indication", MetricDefinition(PrometheusMetricType::counter, "Number of responses sent, packet-cache hits excluded, that were in the Bogus state because a NODATA or NXDOMAIN answer lacked the required SOA and/or NSEC(3) records")},
{"dnssec-result-bogus-signature-not-yet-valid", MetricDefinition(PrometheusMetricType::counter, "Number of responses sent, packet-cache hits excluded, that were in the Bogus state because the signature inception time in the RRSIG was not yet valid")},
{"dnssec-result-bogus-signature-expired", MetricDefinition(PrometheusMetricType::counter, "Number of responses sent, packet-cache hits excluded, that were in the Bogus state because the signature expired time in the RRSIG was in the past")},
{"dnssec-result-bogus-unsupported-dnskey-algo", MetricDefinition(PrometheusMetricType::counter, "Number of responses sent, packet-cache hits excluded, that were in the Bogus state because a DNSKEY RRset contained only unsupported DNSSEC algorithms")},
{"dnssec-result-bogus-unsupported-ds-digest-type", MetricDefinition(PrometheusMetricType::counter, "Number of responses sent, packet-cache hits excluded, that were in the Bogus state because a DS RRset contained only unsupported digest types")},
{"dnssec-result-bogus-no-zone-key-bit-set", MetricDefinition(PrometheusMetricType::counter, "Number of responses sent, packet-cache hits excluded, that were in the Bogus state because no DNSKEY with the Zone Key bit set was found")},
{"dnssec-result-bogus-revoked-dnskey", MetricDefinition(PrometheusMetricType::counter, "Number of responses sent, packet-cache hits excluded, that were in the Bogus state because all DNSKEYs were revoked")},
{"dnssec-result-bogus-invalid-dnskey-protocol", MetricDefinition(PrometheusMetricType::counter, "Number of responses sent, packet-cache hits excluded, that were in the Bogus state because all DNSKEYs had invalid protocols")},
{"x-dnssec-result-bogus", MetricDefinition(PrometheusMetricType::counter, "Same as corresponding metric without ``x-`` prefix, for names in :ref:`setting-yaml-dnssec.x_dnssec_names`.")},
{"x-dnssec-result-bogus-no-valid-dnskey", MetricDefinition(PrometheusMetricType::counter, "Same as corresponding metric without ``x-`` prefix, for names in :ref:`setting-yaml-dnssec.x_dnssec_names`.")},
{"x-dnssec-result-bogus-invalid-denial", MetricDefinition(PrometheusMetricType::counter, "Same as corresponding metric without ``x-`` prefix, for names in :ref:`setting-yaml-dnssec.x_dnssec_names`.")},
{"x-dnssec-result-bogus-unable-to-get-dss", MetricDefinition(PrometheusMetricType::counter, "Same as corresponding metric without ``x-`` prefix, for names in :ref:`setting-yaml-dnssec.x_dnssec_names`.")},
{"x-dnssec-result-bogus-unable-to-get-dnskeys", MetricDefinition(PrometheusMetricType::counter, "Same as corresponding metric without ``x-`` prefix, for names in :ref:`setting-yaml-dnssec.x_dnssec_names`.")},
{"x-dnssec-result-bogus-self-signed-ds", MetricDefinition(PrometheusMetricType::counter, "Same as corresponding metric without ``x-`` prefix, for names in :ref:`setting-yaml-dnssec.x_dnssec_names`.")},
{"x-dnssec-result-bogus-no-rrsig", MetricDefinition(PrometheusMetricType::counter, "Same as corresponding metric without ``x-`` prefix, for names in :ref:`setting-yaml-dnssec.x_dnssec_names`.")},
{"x-dnssec-result-bogus-no-valid-rrsig", MetricDefinition(PrometheusMetricType::counter, "Same as corresponding metric without ``x-`` prefix, for names in :ref:`setting-yaml-dnssec.x_dnssec_names`.")},
{"x-dnssec-result-bogus-missing-negative-indication", MetricDefinition(PrometheusMetricType::counter, "Same as corresponding metric without ``x-`` prefix, for names in :ref:`setting-yaml-dnssec.x_dnssec_names`.")},
{"x-dnssec-result-bogus-signature-not-yet-valid", MetricDefinition(PrometheusMetricType::counter, "Same as corresponding metric without ``x-`` prefix, for names in :ref:`setting-yaml-dnssec.x_dnssec_names`.")},
{"x-dnssec-result-bogus-signature-expired", MetricDefinition(PrometheusMetricType::counter, "Same as corresponding metric without ``x-`` prefix, for names in :ref:`setting-yaml-dnssec.x_dnssec_names`.")},
{"x-dnssec-result-bogus-unsupported-dnskey-algo", MetricDefinition(PrometheusMetricType::counter, "Same as corresponding metric without ``x-`` prefix, for names in :ref:`setting-yaml-dnssec.x_dnssec_names`.")},
{"x-dnssec-result-bogus-unsupported-ds-digest-type", MetricDefinition(PrometheusMetricType::counter, "Same as corresponding metric without ``x-`` prefix, for names in :ref:`setting-yaml-dnssec.x_dnssec_names`.")},
{"x-dnssec-result-bogus-no-zone-key-bit-set", MetricDefinition(PrometheusMetricType::counter, "Same as corresponding metric without ``x-`` prefix, for names in :ref:`setting-yaml-dnssec.x_dnssec_names`.")},
{"x-dnssec-result-bogus-revoked-dnskey", MetricDefinition(PrometheusMetricType::counter, "Same as corresponding metric without ``x-`` prefix, for names in :ref:`setting-yaml-dnssec.x_dnssec_names`.")},
{"x-dnssec-result-bogus-invalid-dnskey-protocol", MetricDefinition(PrometheusMetricType::counter, "Same as corresponding metric without ``x-`` prefix, for names in :ref:`setting-yaml-dnssec.x_dnssec_names`.")},
{"x-dnssec-result-indeterminate", MetricDefinition(PrometheusMetricType::counter, "Same as corresponding metric without ``x-`` prefix, for names in :ref:`setting-yaml-dnssec.x_dnssec_names`.")},
{"x-dnssec-result-nta", MetricDefinition(PrometheusMetricType::counter, "Same as corresponding metric without ``x-`` prefix, for names in :ref:`setting-yaml-dnssec.x_dnssec_names`.")},
{"x-dnssec-result-insecure", MetricDefinition(PrometheusMetricType::counter, "Same as corresponding metric without ``x-`` prefix, for names in :ref:`setting-yaml-dnssec.x_dnssec_names`.")},
{"x-dnssec-result-secure", MetricDefinition(PrometheusMetricType::counter, "Same as corresponding metric without ``x-`` prefix, for names in :ref:`setting-yaml-dnssec.x_dnssec_names`.")},
{"idle-tcpout-connections", MetricDefinition(PrometheusMetricType::gauge, "Number of connections in the TCP idle outgoing connections pool")},
{"source-disallowed-notify", MetricDefinition(PrometheusMetricType::counter, "Number of NOTIFY operations not allowed by allow-notify-from")},
{"zone-disallowed-notify", MetricDefinition(PrometheusMetricType::counter, "Number of NOTIFY operations not allowed by allow-notify-for")},
{"non-resolving-nameserver-entries", MetricDefinition(PrometheusMetricType::gauge, "Number of entries in the non-resolving NS name cache")},
{"maintenance-usec", MetricDefinition(PrometheusMetricType::counter, "Time spent doing internal maintenance, including Lua maintenance")},
{"maintenance-calls", MetricDefinition(PrometheusMetricType::counter, "Number of times internal maintenance has been called, including Lua maintenance")},
{"auth-formerr-answers", MetricDefinition(PrometheusMetricType::multicounter, "Counts the rcodes returned by authoritative servers.")},
{"packetcache-contended", MetricDefinition(PrometheusMetricType::counter, "Number of contended packet cache lock acquisitions")},
{"packetcache-acquired", MetricDefinition(PrometheusMetricType::counter, "Number of packet cache lock acquisitions")},
{"nod-events", MetricDefinition(PrometheusMetricType::counter, "Count of NOD events")},
{"udr-events", MetricDefinition(PrometheusMetricType::counter, "Count of UDR events")},
{"max-chain-length", MetricDefinition(PrometheusMetricType::counter, "Maximum chain length")},
{"max-chain-weight", MetricDefinition(PrometheusMetricType::counter, "Maximum chain weight")},
{"chain-limits", MetricDefinition(PrometheusMetricType::counter, "Chain limits reached")},
{"tcp-overflow", MetricDefinition(PrometheusMetricType::counter, "Incoming TCP limits reached")},
{"ecs-missing", MetricDefinition(PrometheusMetricType::counter, "Number of answers where ECS info was missing")},
{"cookie-malformed", MetricDefinition(PrometheusMetricType::counter, "Number of malformed cookies received")},
{"cookie-matched", MetricDefinition(PrometheusMetricType::counter, "Number of matching cookies received")},
{"cookie-mismatch-tcp", MetricDefinition(PrometheusMetricType::counter, "Number of mismatched cookies received over TCP")},
{"cookie-mismatch-udp", MetricDefinition(PrometheusMetricType::counter, "Number of mismatched cookies received over UDP")},
{"cookie-not-in-reply", MetricDefinition(PrometheusMetricType::counter, "Number of times an authoritative server sent a reply without a cookie")},
{"cookie-retry", MetricDefinition(PrometheusMetricType::counter, "Number of retries because authoritative server sent a BADCOOKIE reply")},
{"cookies-supported", MetricDefinition(PrometheusMetricType::counter, "Number of authoritative server cookie probes resulting in success")},
{"cookies-unsupported", MetricDefinition(PrometheusMetricType::counter, "Number of authoritative server cookie probes not resulting in success")},
{"remote-logger-count-o-0", MetricDefinition(PrometheusMetricType::multicounter, "Number of remote logging events")},
{"cumul-clientanswers-count", MetricDefinition(PrometheusMetricType::histogram, "Cumulative counts of answer times of authoritative servers in buckets less than x microseconds.")},
{"cumul-authanswers-count4", MetricDefinition(PrometheusMetricType::histogram, "Cumulative counts of answer times to clients in buckets less than x microseconds.")},
{"policy-hits", MetricDefinition(PrometheusMetricType::multicounter, "Number of policy decisions based on Lua")},
{"proxy-mapping-total-n-0", MetricDefinition(PrometheusMetricType::multicounter, "Proxy mappings done")},
